"""
Testing suite for aat_wrest
"""

import argparse
import logging
import sys
from typing import List, Tuple
from unittest.mock import MagicMock, patch

import pendulum as pendulum
import pytest
from aat_wrest.metadata_wrester import WrestWorkflowMetadata

logger = logging.getLogger("test_aat_wrest")
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))


args_cal = argparse.Namespace(
    stdcals="uid://evla/execblock/125dac14-c53b-4489-9c43-7d6e2739ec32", stdimg=None
)
result_cal = (
    '{"sdmId": "17B-197.sb34663512.eb34806505.58108.78427738426", "projectCode": "17B-197", '
    '"title": "Radio-Active CoWS: Extended FIRST Sources in MaDCoWS Clusters", '
    '"startTime": 58108.7842916667, "observer": "Anthony Gonzalez", "telescope": "EVLA",'
    ' "created_at": 2021-01-01T00:00:00}'
)

args_image = argparse.Namespace(stdcals=None, stdimg="brain_000.58099.67095825232")
result_img = (
    '{"product_locator": "uid://evla/execblock/ec082e65-452d-4fec-ad88-f5b4af1f9e36",'
    ' "projectCode": "Operations", "title": "", "startTime": 58099.6710792824, '
    '"observer": "VLA Operations", "telescope": "EVLA", "created_at": "2021-06-30T20:18:21"}'
)


def mock_wrester(args: argparse.Namespace) -> WrestWorkflowMetadata:
    with patch("psycopg2.connect") as mock_connect:
        if args.stdcals is not None:
            return WrestWorkflowMetadata(connection=mock_connect, spl=args.stdcals)
        if args.stdimg is not None:
            return WrestWorkflowMetadata(connection=mock_connect, sdm_id=args.stdimg)


@pytest.fixture
def fake_obs_metadata() -> List[Tuple[str]]:
    return [
        (
            "uid://evla/execblock/8fbfb54b-d141-42fe-b079-609339a69cba",
            "Ka",
            "D",
            "1234567",
            "123598898",
        )
    ]


cal_wrester = mock_wrester(args_cal)
img_wrester = mock_wrester(args_image)


class TestAatWrest:
    def test_init(self, fake_obs_metadata: List[Tuple[str]]):
        fileset_id = "17B-197.sb34663512.eb34806505.58108.78427738426"
        with patch("psycopg2.connect") as mock_connect:
            mock_connect.cursor.return_value.__enter__.return_value.fetchall.return_value = (
                fake_obs_metadata
            )
            wrester_no_spl = WrestWorkflowMetadata(
                connection=mock_connect, spl=None, fileset_id=fileset_id
            )
            assert wrester_no_spl.spl == "uid://evla/execblock/8fbfb54b-d141-42fe-b079-609339a69cba"

    @patch("json.dumps", MagicMock(return_value=result_cal))
    def test_wrest_standard_cal_info(self):
        cal_wrester.conn.cursor.return_value.fetchall.return_value = [
            "17B-197.sb34663512.eb34806505.58108.78427738426",
            "17B-197",
            "Radio-Active CoWS: Extended FIRST Sources in MaDCoWS Clusters",
            58108.7842916667,
            "Anthony Gonzalez",
            "EVLA",
            str(pendulum.now().in_timezone("UTC").format("YYYY-MM-DDTHH:mm:ss")),
        ]
        assert args_cal.stdcals == "uid://evla/execblock/125dac14-c53b-4489-9c43-7d6e2739ec32"
        value = cal_wrester.wrest_standard_cal_info()
        assert (
            value
            == '{"sdmId": "17B-197.sb34663512.eb34806505.58108.78427738426", "projectCode": "17B-197", '
            '"title": "Radio-Active CoWS: Extended FIRST Sources in MaDCoWS Clusters", '
            '"startTime": 58108.7842916667, "observer": "Anthony Gonzalez", "telescope": "EVLA",'
            ' "created_at": 2021-01-01T00:00:00}'
        )

    @patch("json.dumps", MagicMock(return_value=result_img))
    def test_wrest_standard_image_info(self):
        img_wrester.conn.cursor.return_value.fetchall.return_value = [
            "uid://evla/execblock/ec082e65-452d-4fec-ad88-f5b4af1f9e36",
            "Operations",
            "",
            58099.6710792824,
            "VLA Operations",
            "EVLA",
            "2021-06-30T20:18:21",
        ]
        assert args_image.stdimg == "brain_000.58099.67095825232"
        value = img_wrester.wrest_standard_image_info()
        assert (
            value
            == '{"product_locator": "uid://evla/execblock/ec082e65-452d-4fec-ad88-f5b4af1f9e36",'
            ' "projectCode": "Operations", "title": "", "startTime": 58099.6710792824, '
            '"observer": "VLA Operations", "telescope": "EVLA", "created_at": "2021-06-30T20:18:21"}'
        )

    def test_wrest_obs_metadata_from_fileset_id(self, fake_obs_metadata: List[Tuple[str]]):
        cal_wrester.conn.cursor.return_value.__enter__.return_value.fetchall.return_value = (
            fake_obs_metadata
        )
        metadata = cal_wrester.wrest_obs_metadata_from_fileset_id(
            "17B-197.sb34663512.eb34806505.58108.78427738426"
        )

        assert metadata == {
            "spl": "uid://evla/execblock/8fbfb54b-d141-42fe-b079-609339a69cba",
            "bands": "Ka",
            "array_config": "D",
            "obs_start_time": "1234567",
            "obs_end_time": "123598898",
        }
