"""
Tests for Messenger's helper functions and messages.
"""
import logging
import sys
from unittest.mock import MagicMock, patch

import pytest
from kombu import Exchange, Queue

from messaging.messenger import AMQPServerInitializer, Messenger

logger = logging.getLogger("messenger")
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))

"""
Tests for Messenger's helper functions
"""


def test_setup_connection(test_server_initializer: AMQPServerInitializer):
    """
    Tests that a Connection object is initialized; the object is mocked
    """
    mock_connection = test_server_initializer.connection
    assert mock_connection
    assert hasattr(mock_connection, "hostname")
    assert hasattr(mock_connection, "userid")
    assert hasattr(mock_connection, "port")


def test_declare_exchange(test_server_initializer: AMQPServerInitializer):
    """
    Tests that an exchange is properly declared
    """
    test_exchange = test_server_initializer.create_exchange()
    assert isinstance(test_exchange, Exchange)
    assert test_exchange.name == "workspaces"
    assert test_exchange.type == "topic"


def test_create_queues(test_server_initializer: AMQPServerInitializer):
    """
    Tests that queues are properly created
    """
    original_setup_queues = test_server_initializer.setup_queues
    test_server_initializer.setup_queues = MagicMock()

    queues = test_server_initializer.create_queues(Exchange("test"))
    assert len(queues) == 0

    # Test service_name = "workflow"
    test_server_initializer.service_name = "workflow"
    queues = test_server_initializer.create_queues(Exchange("test"))
    test_server_initializer.setup_queues.assert_called_with(queues)
    assert len(queues) == 1
    (queue,) = queues
    assert queue.name == "workflow_queue"
    assert queue.routing_key == "workflow"

    # Test service_name = "capability"
    test_server_initializer.service_name = "capability"
    queues = test_server_initializer.create_queues(Exchange("cp_test"))
    test_server_initializer.setup_queues.assert_called_with(queues)
    assert len(queues) == 2

    cp_queue, cp_wf_queue = queues
    assert cp_queue.name == "capability_queue"
    assert cp_wf_queue.name == "cp_wf_queue"
    assert cp_queue.routing_key == "capability"
    assert cp_wf_queue.routing_key == "workflow"

    # Test service_name = "archive"
    test_server_initializer.service_name = "archive"
    queues = test_server_initializer.create_queues(Exchange("archive.events_test"))
    test_server_initializer.setup_queues.assert_called_with(queues)
    (archive_queue,) = queues
    assert len(queues) == 1
    assert archive_queue.name == "ws_ingestion_complete_queue"
    assert archive_queue.routing_key == "ingestion-complete.#"

    test_server_initializer.setup_queues = original_setup_queues


@patch("messaging.messenger.Queue.maybe_bind")
@patch("messaging.messenger.Queue.declare")
def test_setup_queues(
    declare: MagicMock, maybe_bind: MagicMock, test_server_initializer: AMQPServerInitializer
):
    """
    Tests that queues are properly set up
    """

    maybe_bind_calls = maybe_bind.call_count
    declare_calls = declare.call_count
    test_exchange = Exchange("test")
    test_queues = [Queue("test", exchange=test_exchange, routing_key="test")]
    test_server_initializer.setup_queues(test_queues)

    assert maybe_bind.call_count == maybe_bind_calls + 2
    assert declare.call_count == declare_calls + 1


"""
Tests for Messenger methods
"""


@pytest.mark.skip("TODO")
def test_receives_message(test_messenger: Messenger):
    """
    Tests that the message is correctly passed on to the destination
    """
    raise NotImplementedError


@pytest.mark.skip("TODO")
def test_gets_consumers(test_messenger: Messenger):

    """
    Tests that a correct list of consumers is generated.

    :param test_messenger: a simple faux Messenger
    :return:
    """
    raise NotImplementedError


@pytest.mark.skip("TODO")
def test_sends_message(test_messenger: Messenger):
    """
    Tests that a message is properly sent (self.producer.publish is called)

    :param test_messenger: a simple faux Messenger
    :return:
    """

    raise NotImplementedError
